# -*- coding: utf-8 -*-

import click
import codecs
import datetime
import gettext
import sys
import tabulate
import os

from migrator.config import Config
from migrator.executor import Executor

# Установка локализации
translation = gettext.translation('migrator_cli', './locale', fallback=True)
if sys.version_info > (3,):
    _ = translation.gettext
    translation.install()
else:
    _ = translation.ugettext
    translation.install(unicode=True)


def choices_prompt(message, choices):
    return click.prompt(
        u'{}\n{}'.format(
            u'\n'.join([u'{} - {}'.format(k, v) for k, v in sorted(choices.items(), key=lambda x: x[0])]),
            message
        ),
        default='0', type=click.Choice(sorted(choices.keys())), show_default=True
    )


def halt(messgae):
    click.echo(messgae)
    sys.exit(1)


def get_one_revision(migrator, rev):
    revisions = migrator.get_migrations_by_hash(rev)
    if not revisions:
        halt(_(u'Revision {} not found.').format(rev))
    if len(revisions) > 1:
        halt(_(u'Revision {} has too much matches ({}).').format(rev, len(revisions)))
    return revisions[0]


@click.group()
@click.option('-c', '--config', type=click.Path(), default='migrator.cfg')
@click.pass_context
def cli(ctx, config):
    config = os.path.realpath(config)
    cfg = Config()
    ctx.obj = ctx.obj or {}
    if os.path.exists(config):
        cfg.load(config)
        ctx.obj['config_path'] = config
    else:
        cfg.make_default()
        ctx.obj['config_path'] = None
    ctx.obj.update({'cfg': cfg})


@cli.command()
@click.option('--type', 'db_type_arg', default=None)
@click.option('--name', 'db_name_arg', default=None)
@click.option('--user', 'db_user_arg', default=None)
@click.option('--password', 'db_password_arg', default=None)
@click.option('--host', 'db_host_arg', default=None)
@click.option('--port', 'db_port_arg', default=None)
@click.option('--dir', 'migrations_dir_arg', default=None)
@click.option('--sys_path', 'sys_path_arg', default=None)
@click.option('--models', 'models_path_arg', default=None)
@click.option('--excluded', 'models_excluded_arg', default=None)
@click.option('--force', default=False, is_flag=True)
@click.pass_context
def create_config(
        ctx, db_type_arg, db_name_arg, db_user_arg, db_password_arg, db_host_arg, db_port_arg,
        migrations_dir_arg, sys_path_arg, models_path_arg, models_excluded_arg, force
):
    config_path = ctx.obj['config_path'] or 'migrator.cfg'
    ask_edit = all(
        x is None for x in (
            db_type_arg, db_name_arg, db_user_arg, db_password_arg, db_host_arg, db_port_arg,
            migrations_dir_arg, sys_path_arg, models_path_arg, models_excluded_arg
        )
    )
    if os.path.exists(config_path):
        if not force and not click.confirm(_(u'Config already exists. Overwrite?'), default=False):
            return
    db_types = {'0': 'postgres', '1': 'sqlite', '2': 'mysql'}

    if db_type_arg is None:
        result = choices_prompt(_(u'Database type'), db_types)
        db_type = db_types[result]
    else:
        db_type = db_type_arg
    click.echo(db_type)
    db_name = (
        click.prompt(_(u'Database'), default='peewee{}'.format('.db' if db_type == 'sqlite' else ''))
        if db_name_arg is None else db_name_arg
    )
    if db_type == 'sqlite':
        db_url = 'sqlite:///{}'.format(db_name)
    else:
        user = click.prompt(_(u'User'), default='peewee') if db_user_arg is None else db_user_arg
        password = click.prompt(_(u'Password'), hide_input=True) if db_password_arg is None else db_password_arg
        host = click.prompt(_(u'Host'), default='127.0.0.1') if db_host_arg is None else db_host_arg
        port = (
            click.prompt(_(u'Port'), default='5432' if db_type == 'postgres' else '3306')
            if db_port_arg is None else db_port_arg
        )
        db_url = '{}://{}:{}@{}:{}/{}'.format(db_type, user, password, host, port, db_name)

    default_path = os.path.abspath('.')
    migrations_dir = (
        click.prompt(
            _(u'Migrations directory'), default=os.path.join(default_path, 'migrations'), type=click.Path(
                exists=True, file_okay=False, writable=True, readable=True, resolve_path=True
            )
        ) if migrations_dir_arg is None else migrations_dir_arg
    )
    sys_path = (
        click.prompt(
            _(u'PYTHON_PATH for project (Use : as delimiter)'), default=default_path, type=click.Path(
                exists=True, file_okay=False, writable=False, readable=True
            )
        ) if sys_path_arg is None else sys_path_arg
    )

    models_path = (
        click.prompt(_(u'Models path (Comma separated)'), default='app.models')
        if models_path_arg is None else models_path_arg
    )

    excluded_models = (
        click.prompt(_(u'Excluded models by name (Comma separated)'), default='')
        if models_excluded_arg is None else models_excluded_arg
    )

    # Тут записать все в новый конфиг
    cfg = Config()
    cfg.update({
        cfg.BASE_SECTION: {
            cfg.MIGRATOR_DB_URL: db_url,
            cfg.MIGRATOR_DB_TYPE: db_type,
            cfg.MIGRATOR_SYS_PATH: sys_path,
            cfg.MIGRATOR_MIGRATIONS_DIR: migrations_dir,
            cfg.MIGRATOR_MODELS_PATH: models_path,
            cfg.MIGRATOR_EXCLUDED_MODELS: excluded_models
        }
    })

    config = cfg.to_string_io().getvalue()

    if ask_edit and click.confirm(_(u'Edit config?'), default=True):
        edited = click.edit(config)
        if edited is not None:  # None - если редактирование отменено
            config = edited
    with codecs.open(config_path, 'w', 'utf-8') as f:
        f.write(config)
    click.echo(_(u'Config successfully saved!'))


@cli.command('make')
@click.option('--from', 'migration_type', default='db', type=click.Choice(['db', 'last', 'rev', 'empty']))
@click.option('--rev', default=None)
@click.option('--name', default=None)
@click.pass_context
def make_migration(ctx, migration_type, rev, name):
    migrator = Executor(ctx.obj['cfg'])
    if rev is None and migration_type == 'rev':
        halt(_(u'--rev param required'))
    migration_name = (click.prompt(_(u'Migration title')) or None) if name is None else name

    if migration_type == 'empty':
        migrator.make_empty_migration(migration_name=migration_name)
    if migration_type == 'db':
        migrator.migrate_from_db(migration_name=migration_name)
        return
    if migration_type == 'last':
        last = sorted(migrator.get_migrations(), key=lambda x: -x['time'])
        if not last:
            halt(_(u'There are no latest migration.'))
        migrator.migrate_from_migration(last[0], migration_name=migration_name)
        return
    if migration_type == 'rev':
        revisions = migrator.get_migrations_by_hash(rev)
        if not revisions:
            click.echo(_(u'Revision {} not found.').format(rev))
        if len(revisions) > 1:
            halt(_(u'Revision {} has too much matches ({}).').format(rev, len(revisions)))
        migrator.migrate_from_migration(revisions[0], migration_name=migration_name)
        return


@cli.command('apply')
@click.option('--force', default=False, is_flag=True)
@click.option('--fake', default=False, is_flag=True)
@click.argument('rev')
@click.pass_context
def apply_migration(ctx, force, fake, rev):
    migrator = Executor(ctx.obj['cfg'])
    revision = get_one_revision(migrator, rev)
    if migrator.check_status(revision['hash']) == migrator.STATUS_APPLIED:
        if not force and not click.confirm(_(u'Migration already applied. Repeat?'), default=False):
            halt(_(u'Abort'))
    if not migrator.check_dependencies(revision):
        not_applied = [x for x in revision['dependencies'] if migrator.check_status(x) != migrator.STATUS_APPLIED]
        halt(_(u'Migration dependencies not applied: {}').format(u','.join(not_applied)))
    migrator.apply(revision, fake=fake)
    click.echo(_(u'Migration {} applied successfully!').format(revision['hash']))


@cli.command('list')
@click.option('--sort', default='time', type=click.Choice(['name', 'time', 'hash', 'status']))
@click.option('--reverse', default=False, is_flag=True)
@click.pass_context
def migrations_list(ctx, sort, reverse):
    migrator = Executor(ctx.obj['cfg'])
    headers = [_(u'Time'), _(u'Title'), _(u'Migration'), _(u'Dependencies'), _(u'Applied'), _(u'Required')]
    rows = []
    for migration in migrator.get_migrations():
        rows.append([
            datetime.datetime.fromtimestamp(migration['time']),
            migration['name'],
            migration['hash'],
            ', '.join(migration['dependencies']) if migration['dependencies'] else _(u'No'),
            _(u'Yes') if migration['status'] == migrator.STATUS_APPLIED else _(u'No'),
            u'{}, {}'.format(_(u'Yes'), migration['required'] + 1) if migration['required'] is not None else _(u'No')
        ])
    sort_column = {'time': 0, 'name': 1, 'hash': 2, 'status': 4}[sort]
    rows = sorted(rows, key=lambda x: x[sort_column])
    if reverse:
        rows = reversed(rows)
    click.echo(tabulate.tabulate(rows, headers, tablefmt='psql'))


@cli.command('require')
@click.option('--after', default=None)
@click.argument('rev')
@click.pass_context
def mark_required(ctx, after, rev):
    migrator = Executor(ctx.obj['cfg'])
    revision = get_one_revision(migrator, rev)['hash']
    if after is not None:
        after = get_one_revision(migrator, after)['hash']
    migrator.make_required(revision, after=after)
    click.echo(_(u'Revision {} marked as required').format(revision))


@cli.command('up')
@click.option('--fake', default=False, is_flag=True)
@click.pass_context
def up_required(ctx, fake):
    migrator = Executor(ctx.obj['cfg'])
    required = migrator.get_required()
    if not required:
        halt(_(u'There are no required migrations (Empty required.json)'))

    to_apply = [rev for rev in required if migrator.check_status(rev) != migrator.STATUS_APPLIED]
    if not to_apply:
        halt(_(u'Project already up to date'))
    for rev in to_apply:
        revision = get_one_revision(migrator, rev)
        if not migrator.check_dependencies(revision):
            not_applied = [x for x in revision['dependencies'] if migrator.check_status(x) != migrator.STATUS_APPLIED]
            halt(_(u'Migration dependencies not applied: {}').format(u','.join(not_applied)))
        migrator.apply(revision, fake=fake)
        click.echo(_(u'Migration {} applied successfully!').format(revision['hash']))


if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import codecs
import hashlib
import os
import sys
import time

import json
from playhouse.migrate import SqliteMigrator, MySQLMigrator, PostgresqlMigrator

from migrator.code_generator import CodeGenerator
from migrator.db_inspector import Inspector
from migrator.collector import ChangesCollector


__all__ = ['Executor']


def extend_path(path, index=0):
    for p in reversed(path.split(':')):
        if p not in sys.path:
            sys.path.insert(index, p)


class Executor(object):
    MIGRATORS = {
        'sqlite': SqliteMigrator,
        'mysql': MySQLMigrator,
        'postgres': PostgresqlMigrator
    }
    CodeGenerator = CodeGenerator
    STATUS_APPLIED = 'applied'
    STATUS_AVAILABLE = 'available'
    REQUIRED_FILE = 'required.json'

    def __init__(self, config):
        self.config = config
        # Необходимо для корректной работы
        extend_path(config.get_setting(config.MIGRATOR_SYS_PATH), 1)
        self.migrations_in_path = False

    def _extend_migrations(self):
        if not self.migrations_in_path:
            extend_path(self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR), 0)
            self.migrations_in_path = True

    def get_migrator(self):
        migrator = self.MIGRATORS.get(
            self.config.get_setting(self.config.MIGRATOR_DB_TYPE), lambda x: None
        )(self.get_db_obj())
        if migrator is None:
            raise Exception
        return migrator

    def import_migration(self, migration):
        return __import__(migration['import'])

    def get_migrations(self):
        migrations_dir = self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR)
        # Необходимо для работы __import__
        self._extend_migrations()

        for f in os.listdir(migrations_dir):
            if f.startswith('migration_') and f.endswith('.py'):
                revision = f.split('migration_', 1)[-1].rsplit('.py', 1)[0]
                yield self.fetch_migration(revision)

    def get_migrations_by_hash(self, base_hash, migrations=None):
        migrations = self.get_migrations() if migrations is None else migrations
        return [x for x in migrations if x['hash'].startswith(base_hash)]

    def fetch_migration(self, revision):
        self._extend_migrations()
        migration = __import__('migration_{}'.format(revision))
        return {
            'time': migration.MIGRATION_TIME,
            'name': migration.MIGRATION_NAME,
            'hash': revision,
            'file': 'migration_{}.py'.format(revision),
            'import': 'migration_{}'.format(revision),
            'dependencies': migration.MIGRATION_DEPENDENCIES,
            # Примененность
            'status': self.check_status(revision),
            # Обязательность
            'required': self.check_required_position(revision)
        }

    def apply(self, migration, fake=False):
        if not fake:
            module = self.import_migration(migration)
            module.up(self.get_db_obj(), self.get_migrator())
        self.make_applied(migration['hash'])

    def get_applied(self):
        migrations_dir = self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR)
        # TODO: Добавить бэкенд сохранения в БД
        try:
            with codecs.open(os.path.join(migrations_dir, '.applied.json'), 'r', 'utf-8') as f:
                return json.loads(f.read())
        except:
            return []

    def get_required(self):
        migrations_dir = self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR)
        try:
            with codecs.open(os.path.join(migrations_dir, self.REQUIRED_FILE), 'r', 'utf-8') as f:
                return json.loads(f.read())
        except:
            return []

    def make_required(self, revision, after=None):
        required = self.get_required()
        position = len(required)  # Последним
        if after is not None:
            try:
                position = required.index(after) + 1
            except ValueError:
                pass
        required.insert(position, revision)
        migrations_dir = self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR)
        with codecs.open(os.path.join(migrations_dir, self.REQUIRED_FILE), 'w', 'utf-8') as f:
            f.write(json.dumps(required))

    def make_applied(self, revision):
        applied = set(self.get_applied())
        applied.add(revision)
        # TODO: Добавить бэкенд сохранения в БД
        migrations_dir = self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR)
        with codecs.open(os.path.join(migrations_dir, '.applied.json'), 'w', 'utf-8') as f:
            f.write(json.dumps(list(applied)))

    def check_status(self, revision):
        applied = self.get_applied()

        return self.STATUS_APPLIED if revision in applied else self.STATUS_AVAILABLE

    def check_required_position(self, revision):
        required = self.get_required()

        try:
            return required.index(revision)
        except ValueError:
            return None

    def check_dependencies(self, migration):
        return all([self.check_status(dep) == self.STATUS_APPLIED for dep in migration['dependencies']])

    def get_db_obj(self):
        return self.config.get_db()

    def check_migrations_package(self):
        init_path = os.path.join(self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR), '__init__.py')
        if not os.path.exists(init_path):
            with codecs.open(init_path, 'w') as f:
                f.write('')

    def make_empty_migration(self, migration_name=None):
        # TODO: сделать генерацию по готовому шаблону, без self.migratemigr
        i = Inspector(models_path=self.config.get_models_paths(), excluded_models=self.config.get_excluded())
        current_models = list(i.inspect_models())
        c = CodeGenerator(current_models)
        imports, models, proxies = c.clses_code()
        old_data = c.clses_json()
        new = {x['name']: x for x in old_data}

        return self.migrate(imports, models, proxies, new, new, old_data, migration_name=migration_name)

    def migrate_from_migration(self, migration=None, migration_name=None):
        # Получение информации о предыдущем состоянии из моделй
        i = Inspector(models_path=[migration['import']], excluded_models=self.config.get_excluded())
        old_models = list(i.inspect_models())
        c = CodeGenerator(old_models)
        old_json = c.clses_json()
        old = {x['name']: x for x in old_json}
        old_data = old_json

        # Получение информации о текущем состоянии из моделей
        i = Inspector(models_path=self.config.get_models_paths(), excluded_models=self.config.get_excluded())
        current_models = list(i.inspect_models())
        c = CodeGenerator(current_models)
        imports, models, proxies = c.clses_code()
        new = {x['name']: x for x in c.clses_json()}

        return self.migrate(
            imports, models, proxies, new, old, old_data, migration_name=migration_name, dependencies=[migration]
        )

    def migrate_from_db(self, migration_name=None):
        # Получение информации о текущем состоянии из моделей
        i = Inspector(models_path=self.config.get_models_paths(), excluded_models=self.config.get_excluded())
        current_models = list(i.inspect_models())
        c = CodeGenerator(current_models)
        imports, models, proxies = c.clses_code()
        new = {x['name']: x for x in c.clses_json()}

        # Получение информации о состоянии из базы данных
        i = Inspector(excluded_models=self.config.get_excluded())
        db_models = list(i.inspect_database(self.get_db_obj()))
        c = CodeGenerator(db_models)
        old_json = c.clses_json()
        old = {x['name']: x for x in old_json}
        old_data = old_json

        return self.migrate(imports, models, proxies, new, old, old_data, migration_name=migration_name)

    def migrate(self, imports, models, proxies, new, old, old_data, migration_name=None, dependencies=None):
        collector = ChangesCollector()
        # Построение таблицы соответствий
        model_matches = collector.get_table_matches(new, old)

        # Операции миграции
        migration = {
            # Базовые операции: создание и удаление таблиц
            'drop': [x for x in old.keys() if x not in model_matches.values()],
            'create': [x for x in new.keys() if x not in model_matches.keys()],
            # Наполняемые данные
            'rename': [],
            'fields': {
                'rename': [],
                'create': [],
                'drop': [],
                'index': [],
                'null': []
            }
        }

        for new_model_name, old_model_name in model_matches.items():
            new_model = new[new_model_name]
            old_model = old[old_model_name]
            # Обработка случая переименования таблицы
            if new_model['table'] != old_model['table']:
                migration['rename'].append((old_model['table'], new_model['table']))

            # Обработка полей
            new_fields = {x['name']: x for x in new[new_model_name]['fields']}
            old_fields = {x['name']: x for x in old[old_model_name]['fields']}

            field_matches = collector.get_field_matches(new[new_model_name]['fields'], new_fields, old_fields)
            # Создание и удаление полей
            migration['fields']['drop'].extend(
                collector.get_columns_to_drop(new_fields, old_fields, new_model, old_model, field_matches)
            )
            migration['fields']['create'].extend(
                collector.get_columns_to_create(new_fields, old_fields, new_model, old_model, field_matches)
            )

            for new_field_name, old_field_name in field_matches.items():
                new_field = new_fields[new_field_name]
                old_field = old_fields[old_field_name]
                # Обработка индексов
                migration['fields']['index'].extend(collector.get_field_indexes(new_field, old_field, new_model, old_model))
                # Обработка null
                migration['fields']['null'].extend(collector.get_field_null(new_field, old_field, new_model, old_model))
                # Обработка переименований полей
                migration['fields']['rename'].extend(collector.get_field_rename(new_field, old_field, new_model, old_model))

        # Генерация кода
        up, down = self.CodeGenerator.changes_code(migration, indent=4)

        # Конечное создание миграции
        return self.make_migration(
            imports, models, old_data, up=up, down=down, migration_name=migration_name, proxies=proxies,
            dependencies=dependencies
        )

    def make_migration(
        self, imports, models, old_data, up=None, down=None, migration_name=None, proxies=None, dependencies=None
    ):

        self.check_migrations_package()

        migration_time = int(time.time())
        migration_hash = hashlib.md5(str(migration_time).encode('utf-8')).hexdigest()

        migration_kwargs = locals()
        migration_kwargs.pop('self', None)

        migration_path = os.path.join(
            self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR),
            'migration_{}.py'.format(migration_hash)
        )
        migration_code = self.CodeGenerator.migration_code(**migration_kwargs)

        with codecs.open(migration_path, 'w', 'utf-8') as f:
            f.write(migration_code)

        return migration_path



# -*- coding: utf-8 -*-
import codecs

import six
from playhouse.db_url import connect
from six.moves import configparser

__all__ = ['Config']


class Config(dict):
    BASE_SECTION = 'migrator'
    MIGRATOR_DB_URL = 'db_url'
    MIGRATOR_DB_TYPE = 'db_type'
    MIGRATOR_SYS_PATH = 'sys_path'
    MIGRATOR_MIGRATIONS_DIR = 'migrations_dir'

    # Список через запятую
    MIGRATOR_MODELS_PATH = 'models_path'
    MIGRATOR_EXCLUDED_MODELS = 'excluded_models'

    def __init__(self, *args, **kwargs):
        self.cp = configparser.ConfigParser()
        super(Config, self).__init__(*args, **kwargs)

    def make_default(self):
        self.clear()
        self.update({
            self.BASE_SECTION: {
                self.MIGRATOR_DB_URL: 'sqlite:///sqlite.db',
                self.MIGRATOR_DB_TYPE: 'sqlite',
                self.MIGRATOR_SYS_PATH: 'project',
                self.MIGRATOR_MIGRATIONS_DIR: 'migrations',
                self.MIGRATOR_MODELS_PATH: 'app.models',
                self.MIGRATOR_EXCLUDED_MODELS: ''
            }
        })
        self.save_to_cp()

    def load(self, file_path):
        self.cp = configparser.ConfigParser()
        self.cp.read(file_path)
        self.load_from_cp()

    def load_from_cp(self):
        for section in self.cp.sections():
            for option in self.cp.options(section):
                value = self.cp.get(section, option)
                self.setdefault(section, {})[option] = value

    def save(self, file_path):
        self.save_to_cp()
        self.cp.write(codecs.open(file_path, 'w', 'utf-8'))

    def to_string_io(self):
        string = six.StringIO()
        self.save_to_cp()
        self.cp.write(string)
        return string

    def save_to_cp(self):
        self.cp = configparser.ConfigParser()
        for section, variables in self.items():
            self.cp.add_section(section)
            for name, value in variables.items():
                self.cp.set(section, name, value)

    def get_db(self):
        url = self.get(self.BASE_SECTION, {}).get(self.MIGRATOR_DB_URL, None)
        if url is None:
            return None
        return connect(url)

    def get_setting(self, key, default=None):
        return self.get(self.BASE_SECTION, {}).get(key, default)

    def _get_list_by_comma(self, value):
        return [x.strip() for x in value.split(',')]

    def get_excluded(self):
        return self._get_list_by_comma(self.get_setting(self.MIGRATOR_EXCLUDED_MODELS, ''))

    def get_models_paths(self):
        return self._get_list_by_comma(self.get_setting(self.MIGRATOR_MODELS_PATH))


# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import json


__all__ = ['CodeGenerator']


class CodeGenerator(object):
    PROXY_PREFIX = 'MigratorProxy__'
    MIGRATION_TEMPLATE = '''# -*- coding: utf-8 -*-
{imports}
from playhouse.migrate import migrate


MIGRATION_NAME = {migration_name}
MIGRATION_TIME = {migration_time}
MIGRATION_DEPENDENCIES = {migration_dependencies}


proxy_db = peewee.Proxy(){proxy_fields}


class BaseModel(peewee.Model):
    class Meta:
        database = proxy_db


{models}{proxy_fields_init}


def up(db, migrator):
    proxy_db.initialize(db)
{up}


def down(db, migrator):
    proxy_db.initialize(db)
{down}
'''

    def __init__(self, models):
        self.models = models

    class Builder(object):
        def __init__(self, indent=0, indent_step=4):
            self.indent = indent
            self.indent_step = indent_step
            self.lines = []

        def tab(self):
            self.indent += self.indent_step
            return self

        def un_tab(self):
            self.indent -= self.indent_step
            return self

        def write_line(self, text):
            self.lines.append('{}{}'.format(' ' * self.indent, text))
            return self

        @property
        def code(self):
            return '\n'.join(self.lines)

    @classmethod
    def _join_lines_with_indent(cls, lines, indent=4):
        return '\n'.join(['{}{}'.format(' ' * indent, x) for x in lines])

    @classmethod
    def _get_import_by_path(cls, path):
        return '.'.join(path.split('.')[:-1])

    def clses_code(self):
        builder = self.Builder()
        to_import = {'peewee'}
        models_str = []
        proxies_str = set()
        for model, db_table, fields in self.models:
            builder.write_line('class {}(BaseModel):'.format(model))
            for field in fields:
                to_import.add(self._get_import_by_path(field['path']))
                params = []
                if field['params']['index'] and not field['params']['unique']:
                    params.append('index=True')
                if field['params']['unique']:
                    params.append('unique=True')
                if field['params']['null']:
                    params.append('null=True')
                if field['params'].get('default'):
                    field_default = field['params']['default']
                    need_import = field_default.get('import') is not None
                    params.append('default={}'.format(
                        field_default['value'] if need_import else field_default['value'].__repr__()
                    ))
                    if need_import:
                        to_import.add(field_default['import'])
                db_field = field['params'].get('initial_kwargs', {}).get('db_field', field['column'])
                if db_field and db_field != field['name']:
                    params.append('db_column={}'.format(db_field.__repr__()))
                to_field = field['params'].get('initial_kwargs', {}).get('to_field', {})
                if to_field and to_field['column'] != 'id':
                    params.append('to_field={}'.format(to_field['column'].__repr__()))
                if field['params'].get('initial_kwargs'):
                    rel = field['params']['initial_kwargs'].get('rel_model', None)
                    if rel:
                        if rel == model:
                            rel = "'self'"
                        else:
                            proxies_str.add(rel)
                            rel = '{}{}'.format(self.PROXY_PREFIX, rel)
                        params.insert(0, rel)
                        # TODO: Попробовать разобраться здесь, добавляя только данные, которые можно сериализовать
                        # for k, v in field['params']['initial_kwargs'].iteritems():
                        #     params.append('{}={}'.format(k, v))
                field_code = '{} = {}({})'.format(field['name'], field['path'], ', '.join(params))
                builder.tab().write_line(field_code).un_tab()
            builder.write_line('').tab().write_line('class Meta:')
            builder.tab().write_line('db_table = {}'.format(db_table.__repr__())).un_tab().un_tab()
            builder.write_line('').write_line('')
        models_str.append(builder.code)
        imports_str = ['import {}'.format(x) for x in to_import]
        return imports_str, models_str, sorted(proxies_str)

    def clses_json(self):
        models_json = []
        for model, db_table, fields in self.models:
            model_json = {'name': model}
            fields_json = []
            imports = set()
            for field in fields:
                import_path = self._get_import_by_path(field['path'])
                params = {}
                if field['params']['index'] and not field['params']['unique']:
                    params['index'] = True
                if field['params']['unique']:
                    params['unique'] = True
                if field['params']['null']:
                    params['null'] = True
                db_field = field['params'].get('initial_kwargs', {}).get('db_field', field['column'])
                if db_field and db_field != field['name']:
                    params['db_field'] = db_field
                if field['params'].get('initial_kwargs'):
                    rel = field['params']['initial_kwargs'].get('rel_model', None)
                    if rel:
                        if rel == model:
                            rel = "'self'"
                        params['__args'] = rel
                field_json = {
                    'path': field['path'], 'params': params, 'name': field['name']
                }
                if import_path != 'peewee':
                    field_json['import'] = import_path
                    imports.add(import_path)
                fields_json.append(field_json)
            model_json.update({'table': db_table, 'fields': fields_json})
            if imports:
                model_json['imports'] = list(imports)
            models_json.append(model_json)
        return sorted(models_json, key=lambda x: x['name'])

    @classmethod
    def changes_code(cls, migration, indent=0):
        builder = cls.Builder(indent=indent)
        up, down = [], ['pass']

        # Создание таблиц
        if migration['create']:
            builder.write_line('').write_line('# Tables creation')
            for model in migration['create']:
                builder.write_line('{model}.create_table()'.format(model=model))

        # Удаление таблиц
        if migration['drop']:
            builder.write_line('').write_line('# Tables deletion')
            for num, db_table in enumerate(migration['drop'], 1):
                if num > 1:
                    builder.write_line('')
                builder.write_line('class ToDrop{num}(BaseModel):'.format(num=num)).tab()
                builder.write_line('class Meta:').tab()
                builder.write_line('db_table = {table}'.format(table=db_table.__repr__())).un_tab().un_tab()
                builder.write_line('')
                builder.write_line('ToDrop{num}.drop_table()'.format(num=num))

        # Переименование таблиц
        pass

        # Создание полей
        if migration['fields']['create']:
            builder.write_line('').write_line('# Fields creation')
            builder.write_line('migrate(').tab()
            for table, db_field, field in migration['fields']['create']:
                builder.write_line(
                    'migrator.add_column({}, {}, {}),'.format(table.__repr__(), db_field.__repr__(), field)
                )

            builder.un_tab().write_line(')')

        # Удаление полей
        if migration['fields']['drop']:
            builder.write_line('').write_line('# Drop fields')
            builder.write_line('migrate(').tab()
            for table, db_field in migration['fields']['drop']:
                builder.write_line('migrator.drop_column({}, {}),'.format(table.__repr__(), db_field.__repr__()))
            builder.un_tab().write_line(')')

        # Переименование полей
        pass

        # Удаление индексов
        to_drop = [x for x in migration['fields']['index'] if x[0] == 'drop']
        if to_drop:
            builder.write_line('').write_line('# Drop fields')
            builder.write_line('migrate(').tab()
            for action, table, db_field in to_drop:
                builder.write_line('migrator.drop_index({}, {}),'.format(table.__repr__(), db_field.__repr__()))
            builder.un_tab().write_line(')')

        # Создание индексов
        to_create = [x for x in migration['fields']['index'] if x[0] == 'add']
        unique = [x for x in to_create if x[-1]]
        index = [x for x in to_create if not x[-1]]
        if index:
            builder.write_line('').write_line('# Create indexes')
            builder.write_line('migrate(').tab()
            for action, table, db_field, is_uniq in index:
                builder.write_line(
                    'migrator.add_index({}, ({},), False),'.format(table.__repr__(), db_field.__repr__())
                )
            builder.un_tab().write_line(')')
        if unique:
            builder.write_line('').write_line('# Create unique indexes')
            builder.write_line('migrate(').tab()
            for action, table, db_field, is_uniq in index:
                builder.write_line(
                    'migrator.add_index({}, ({},), True)),'.format(table.__repr__(), db_field.__repr__())
                )
            builder.un_tab().write_line(')')

        up = [builder.code]

        return up, down

    @classmethod
    def migration_code(
            cls, imports, models, old_data, up=None, down=None, migration_name=None,
            proxies=None, dependencies=None, migration_time=None, migration_hash=None
    ):
        if migration_name is None:
            migration_name = 'Миграция {}'.format(migration_hash)
        if dependencies is None:
            migration_dependencies = []
        else:
            migration_dependencies = '[{}]'.format(', '.join([x['hash'].__repr__() for x in dependencies]))
        if up is None:
            up = ['pass']
        if down is None:
            down = ['pass']
        up = cls._join_lines_with_indent(up)
        down = cls._join_lines_with_indent(down)

        # Обработка прокси
        proxy_fields = []
        proxy_fields_init = []
        for proxy in (proxies or []):
            proxy_fields.append('{}{} = peewee.DeferredRelation()'.format(cls.PROXY_PREFIX, proxy))
            proxy_fields_init.append('{prefix}{proxy}.set_model({proxy})'.format(prefix=cls.PROXY_PREFIX, proxy=proxy))
        proxy_fields = '\n{}'.format('\n'.join(proxy_fields)) if proxy_fields else ''
        proxy_fields_init = '\n{}'.format('\n'.join(proxy_fields_init)) if proxy_fields_init else ''

        return cls.MIGRATION_TEMPLATE.format(
            up=up, down=down, imports='\n'.join(imports), models='\n\n\n'.join(models),
            old_data=json.dumps(old_data, indent=4), migration_name=migration_name.__repr__(),
            migration_time=migration_time, proxy_fields=proxy_fields, proxy_fields_init=proxy_fields_init,
            migration_dependencies=migration_dependencies
        )



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import inspect

import six
import sys
from inspect import getmodule, ismethod, isclass

from peewee import BaseModel, ForeignKeyField
from playhouse.reflection import Introspector

if sys.version_info > (3,):
    if sys.version_info > (3, 4):
        from importlib import reload
    else:
        from imp import reload

__all__ = ['Inspector']


class Inspector(object):
    EXCLUDED_FIELDS = {
        'c', 'choices', 'help_text', 'model_class', 'verbose_name', 'related_name', 'coerce', 'db_column',
        'name', 'db_field', 'deferred'
    }
    NOT_INCLUDE_IF_NULL = {'sequence', 'default', 'constraints', 'schema', 'on_delete', 'on_update', 'extra'}
    NOT_INCLUDE_IF_FALSE = {'primary_key'}
    models_path = ('app.models',)

    def __init__(self, models_path=None, excluded_models=None):
        if models_path is not None:
            self.models_path = models_path
        self.excluded_models = [] if excluded_models is None else excluded_models

    @staticmethod
    def is_model(obj):
        return isclass(obj) and (issubclass(obj, BaseModel) or isinstance(obj, BaseModel))

    def get_models(self):
        for path in self.models_path:
            module = __import__(path, fromlist=['*'])
            models = reload(module)
            for model in dir(models):
                if model[0].isupper() and model[0] != '_':
                    obj = getattr(models, model)
                    if self.is_model(obj) and model not in self.excluded_models:
                        yield model, obj

    def get_database_models(self, db=None):
        i = Introspector.from_database(db)
        return i.generate_models().values()

    def inspect_models(self):
        for model, model_obj in self.get_models():
            yield self.inspect_model(model_obj, model=model)

    def inspect_database(self, db):
        models = self.get_database_models(db)
        for model_obj in models:
            yield self.inspect_model(model_obj)

    def inspect_model(self, model_obj, model=None):
        if model is None:
            model = model_obj.__name__
        fields = list(self.collect_fields(model_obj._meta.sorted_fields, model=model))
        return model, model_obj._meta.db_table, fields

    def get_field_default(self, value):
        primitive_types = tuple(set(
            six.integer_types + six.string_types + (six.binary_type,) + (six.text_type,) + (bool,)
        ))
        if isinstance(value, primitive_types):
            # Простой тип, записывается inline
            return {'value': value}

        if not callable(value):
            # Возможно какой-то мутабельный тип, типа list или dict, такой автоматически не добавляем
            return None
        _module = inspect.getmodule(value)
        if _module is not None:
            # Обычные функции
            return {'value': '{}.{}'.format(_module, value.__name__), 'import': _module}
        # TODO: Надо научиться как-то отлавливать такие builtins, как datetime.datetime.now, у которых module None
        return None

    def collect_fields(self, sorted_fields, model=''):
        for field in sorted_fields:
            name = field.name
            field_path = field.__class__.__name__
            field_module = getmodule(field.__class__)
            if field_module is not None:
                field_module = field_module.__name__
                field_path = '{}.{}'.format(field_module, field_path)

            default_value = self.get_field_default(field.default)
            field_params = {
                'index': field.index, 'unique': field.unique, 'null': field.null,
                'initial_kwargs': {}
            }
            if default_value is not None:
                field_params['default'] = default_value

            for param in dir(field):
                if param[0] == '_' or param in field_params or param in self.EXCLUDED_FIELDS:
                    continue
                try:
                    param_obj = getattr(field, param)
                except:
                    param_obj = None
                if ismethod(param_obj) or (param_obj is None and param in self.NOT_INCLUDE_IF_NULL):
                    continue
                if param_obj is False and param in self.NOT_INCLUDE_IF_FALSE:
                    continue

                field_params['initial_kwargs'][param] = param_obj
            # if field.name != field.db_column:
            #     field_params['initial_kwargs']['db_field'] = field.db_column

            db_column = field.db_column

            if isinstance(field, ForeignKeyField):
                # Сохраняем только название модели
                field_params['initial_kwargs']['rel_model'] = field.rel_model.__name__
                # Нормализация обращения к полю
                db_column = field.name if field.db_column == '{}_id'.format(field.name) else field.db_column
                to_field = {
                    'name': field.to_field.name, 'column': field.to_field.db_column
                }
                if to_field['name'] == 'id' and to_field['column'] == 'id':
                    field_params['initial_kwargs'].pop('to_field', None)
                else:
                    field_params['initial_kwargs']['to_field'] = to_field
                    # Убираем умолчательное значение related_name
                    # model_set = u'{}_set'.format(re.sub('[^\w]+', '_', model.lower()))
                    # if field_params['initial_kwargs']['related_name'] == model_set:
                    #     del field_params['initial_kwargs']['related_name']

            if not field_params['initial_kwargs']:
                field_params.pop('initial_kwargs', None)

            yield {'name': name, 'column': db_column, 'path': field_path, 'params': field_params}



# -*- coding: utf-8 -*-


__all__ = ['ChangesCollector']


class ChangesCollector(object):
    @classmethod
    def get_table_matches(cls, new, old):
        matches = {}
        for name, model in new.items():
            match = None
            if name in old:
                match = name
            else:
                for o_name, o_model in old.items():
                    if o_model['table'] == model['table']:
                        match = o_name
            if match is not None:
                matches[name] = match
        return matches

    @classmethod
    def get_columns_to_drop(cls, new_fields, old_fields, new_model, old_model, field_matches):
        fields_to_drop = []
        matched_old_fields = set(field_matches.values())
        for x in old_fields.keys():
            if x in matched_old_fields:
                continue
            fields_to_drop.append((new_model['table'], cls.get_db_field(old_fields[x])))
        return fields_to_drop

    @classmethod
    def get_columns_to_create(cls, new_fields, old_fields, new_model, old_model, field_matches):
        fields_to_create = []
        existed_fields = set(field_matches.keys())
        for x in new_fields.keys():
            if x in existed_fields:
                continue
            fields_to_create.append(
                (new_model['table'], cls.get_db_field(new_fields[x]), '{}.{}'.format(new_model['name'], x))
            )

        return fields_to_create

    @classmethod
    def get_field_matches(cls, new_ordered, new_fields, old_fields):
        field_matches = {}
        for field in new_ordered:
            match = None
            if field['name'] in old_fields:
                match = field['name']
            # else:
            if True:
                for o_field_name, o_field in old_fields.items():
                    if cls.get_db_field(field) == cls.get_db_field(o_field):
                        match = o_field_name
            if match is not None:
                field_matches[field['name']] = match
        return field_matches

    @classmethod
    def get_db_field(cls, field):
        return str(field.get('params', {}).get('db_field', field['name']))

    @classmethod
    def get_field_indexes(cls, new_field, old_field, new_model, old_model):
        indexes_diff = []
        # Обычные индексы
        if new_field.get('index', False):
            if not old_field.get('index', False):
                indexes_diff.append(('add', new_model['table'], cls.get_db_field(new_field), False))
        else:
            if old_field.get('index', False):
                indexes_diff.append(('drop', new_model['table'], cls.get_db_field(new_field)))
        # Уникальные индексы
        if new_field.get('unique', False):
            if not old_field.get('unique', False):
                indexes_diff.append(('add', new_model['table'], cls.get_db_field(new_field), True))
        else:
            if old_field.get('unique', False):
                indexes_diff.append(('drop', new_model['table'], cls.get_db_field(new_field)))
        return indexes_diff

    @classmethod
    def get_field_null(cls, new_field, old_field, new_model, old_model):
        null_diff = []
        if new_field.get('null', False):
            if not old_field.get('null', False):
                null_diff.append(('add', new_model['table'], cls.get_db_field(new_field)))
        else:
            if old_field.get('null', False):
                null_diff.append((new_model['table'], cls.get_db_field(new_field)))
        return null_diff

    @classmethod
    def get_field_rename(cls, new_field, old_field, new_model, old_model):
        # TODO: Обрабатывать переименование полей модели
        return []



# -*- coding: utf-8 -*-
import codecs
from setuptools import setup, find_packages
import os

here = os.path.abspath(os.path.dirname(__file__))
with codecs.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()
with codecs.open(os.path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    requirements = [x.strip() for x in f.readlines()]

setup(
    name='peewee-migrator',
    version='0.1.1',
    description='Basic migrations support for peewee ORM',
    author='Suguby',
    author_email='suguby@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Natural Language :: Russian',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    url='https://github.com/suguby/peewee-migrator',
    keywords='peewee migrations orm',
    data_files=[
        ('', ['requirements.txt', 'README.rst', ], ),
        ('migrator/locale/ru/LC_MESSAGES', ['migrator/locale/ru/LC_MESSAGES/migrator_cli.mo'], ),
    ],
    packages=find_packages(include=['migrator']),
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'migrator=migrator.cli:cli'
        ]
    }
)
